'' ******************************************************************************
'' * DS3231 Object                                                              *
'' * Tim Moore May 08                                                           *
'' *    Fixed May 09 - status and date/days read incorrectly                    *
'' * James Burrows Oct 07                                                       *
'' * Version 2.0                                                                *
'' ******************************************************************************
''
'' This object provides and example of use of the DS3231 i2c real time clock (RTC)
'' See - for reference:   http://www.maxim-ic.com/quick_view2.cfm/qv_pk/2688
''
'' this object provides the PUBLIC functions:
''  -> settime - sets the clock time
''  -> setdate - sets the date
''  -> gettime - gets the time into the object - use getday, getmonth etc to read the variables
''  -> getdate - gets the date into the object - see above
''  -> getday/getmonth etc - returns the data got when you call gettime/getdate
''
'' this object provides the PRIVATE functions:
''  -> i2c2bcd - performs integer to BCD conversion
''  -> bcd2int - performs BCD to integer conversion
''
'' this object uses the following sub OBJECTS:
''  -> i2cObject
''
'' Revision History:
''  -> V2 - re-Release
''
'' The default address is %1101_0000

CON

VAR
  long  DS3231_Seconds
  long  DS3231_Minutes
  long  DS3231_Hours
  long  DS3231_Date
  long  DS3231_Days
  long  DS3231_Months    
  long  DS3231_Years 
  long  DS3231_Status
   
OBJ
  i2cObject     : "basic_i2c_driver"

PUB setTime(i2cSCL, _deviceAddress, ds_hour, ds_minute, ds_seconds)
  ' set the time
  i2cObject.Start (i2cSCL)
  i2cObject.Write (i2cSCL, _deviceAddress | 0)
  i2cObject.write (i2cSCL, 0)
  DS3231_Seconds := int2bcd(ds_seconds)
  DS3231_Minutes := int2bcd(ds_minute)     
  DS3231_Hours   := int2bcd(ds_hour)
  i2cObject.Write (i2cSCL, DS3231_Seconds)
  i2cObject.Write (i2cSCL, DS3231_Minutes)
  i2cObject.Write (i2cSCL, DS3231_Hours)        
  i2cObject.Stop(i2cSCL)
  DS3231_Status &= $7f                                                          'when set time, reset status flag
  i2cObject.Start (i2cSCL)
  i2cObject.Write (i2cSCL, _deviceAddress | 0)
  i2cObject.write (i2cSCL, $f)
  i2cObject.Write (i2cSCL, DS3231_Status)
  i2cObject.Stop(i2cSCL)

PUB getTime(i2cSCL,_deviceAddress) : ds_seconds
  ' get the time bytes from the clock
  i2cObject.Start (i2cSCL)
  i2cObject.Write (i2cSCL,_deviceAddress | 0)
  i2cObject.Write (i2cSCL,0)
  i2cObject.Start (i2cSCL)
  i2cObject.Write (i2cSCL,_deviceAddress | 1)
  DS3231_Seconds := i2cObject.Read(i2cSCL, i2cObject#ACK)
  DS3231_Minutes := i2cObject.Read(i2cSCL, i2cObject#ACK)
  DS3231_Hours   := i2cObject.Read(i2cSCL, i2cObject#NAK)
  i2cObject.Stop(i2cSCL)
  i2cObject.Start (i2cSCL)                                                      ' read clock status flag
  i2cObject.Write (i2cSCL,_deviceAddress | 0)
  i2cObject.Write (i2cSCL,$f)
  i2cObject.Start (i2cSCL)
  i2cObject.Write (i2cSCL,_deviceAddress | 1)
  DS3231_Status   := i2cObject.Read(i2cSCL, i2cObject#NAK)
  i2cObject.Stop(i2cSCL)
  return DS3231_Status  

  
PUB setDate(i2cSCL, _deviceAddress, ds_day, ds_date, ds_month, ds_year)
  ' set the date 
  i2cObject.Start (i2cSCL)
  i2cObject.Write (i2cSCL,_deviceAddress | 0)
  i2cObject.Write (i2cSCL, 3)
  DS3231_Days   := int2bcd(ds_day)
  DS3231_Date   := int2bcd(ds_date)
  DS3231_Months := int2bcd(ds_month)
  DS3231_Years  := int2bcd(ds_year)      
  i2cObject.Write (i2cSCL,DS3231_Days)  
  i2cObject.Write (i2cSCL,DS3231_Date)
  i2cObject.Write (i2cSCL,DS3231_Months) 
  i2cObject.Write (i2cSCL,DS3231_Years) 
  i2cObject.Stop(i2cSCL)    

  
PUB getDate(i2cSCL,_deviceAddress) : ds_seconds
  ' get the date bytes from the clock
  i2cObject.Start (i2cSCL)
  i2cObject.Write (i2cSCL,_deviceAddress | 0)
  i2cObject.Write (i2cSCL,3)
  i2cObject.Start (i2cSCL)
  i2cObject.Write (i2cSCL,_deviceAddress | 1)
  DS3231_Days   := i2cObject.Read(i2cSCL,i2cObject#ACK)
  DS3231_Date   := i2cObject.Read(i2cSCL,i2cObject#ACK)
  DS3231_Months := i2cObject.Read(i2cSCL,i2cObject#ACK)
  DS3231_Years  := i2cObject.Read(i2cSCL,i2cObject#NAK)
  i2cObject.Stop(i2cSCL)
  
PUB getHours : result
  return bcd2int(DS3231_Hours)

PUB getMinutes : result
  return bcd2int(DS3231_Minutes)

PUB getSeconds : result
  return bcd2int(DS3231_Seconds)  

PUB getDays : result
  return bcd2int(DS3231_Days)

PUB getDayMon : result
  return bcd2int(DS3231_Date)

PUB getMonths : result
  return bcd2int(DS3231_Months)

PUB getYears : result
  return bcd2int(DS3231_Years)      

PUB getStatus : result
  return DS3231_Status
  
pri int2bcd(value) : result
  ' convert integer to BCD (Binary Coded Decimal)
  result := ((value / 10) *16) + (value // 10) 
  return result

pri bcd2int(value) : result
  ' convert BCD (Binary Coded Decimal) to Integer
  result :=((value / 16) *10) + (value // 16) 
  return result
   